import os
from os.path import isfile
import sqlite3
import sqlalchemy
from sqlalchemy import create_engine
import pandas as pd


class CSVDataStore:
    def __init__(self, filename):
        self.filename = filename
        
    def clean(self):
        if os.path.isfile(self.filename):
            os.remove(self.filename)

    def read(self):
        if self.filename:
            try:
                with open(self.filename, 'r') as f:
                    try:
                        df = pd.read_csv(self.filename)
                    except pd.errors.EmptyDataError:
                        pass
                    else:
                        return df
            except FileNotFoundError:
                pass
            
        return pd.DataFrame()

    def save(self, df):
        if isinstance(df, pd.DataFrame) and not df.empty:
            if isfile(self.filename):
                mode = 'a'
                writeHeader = False
            else:
                mode = 'w'
                writeHeader = True

            with open(self.filename, mode) as f:
                df.to_csv(f,header=writeHeader,index=False)


        
class SqliteDataStore:
    def __init__(self, db_name, table_name):
        self.db_name = db_name
        self.table_name = table_name
        
        connection = sqlite3.connect(self.db_name)
        cursor = connection.cursor()
        sql = 'CREATE TABLE IF NOT EXISTS ' + self.table_name + ' (id INT, author INT, author_name TEXT, date text, title TEXT, link TEXT, content TEXT)'
        cursor.execute(sql)
        connection.commit()
        connection.close()
        
    def clean(self):
        connection = sqlite3.connect(self.db_name)
        cursor = connection.cursor()
        sql = 'DROP TABLE IF EXISTS ' + self.table_name
        cursor.execute(sql)
        connection.commit()
        connection.close()
        
    
    def read(self):
        engine = sqlalchemy.create_engine('sqlite:///' + self.db_name, echo=False)
        try:
            df = pd.read_sql(self.table_name, con=engine)
        except sqlalchemy.exc.OperationalError:
            df = pd.DataFrame()
        finally:
            return df
            
        
    def save(self, df):
        if isinstance(df, pd.DataFrame) and not df.empty:
            engine = sqlalchemy.create_engine('sqlite:///' + self.db_name, echo=False)
            df.to_sql(self.table_name, con=engine, if_exists='append',index=False)

            connection = sqlite3.connect(self.db_name)
            connection.commit()
            connection.close()

        


import requests 
import ssl
import re
import pandas as pd


class WPScraper:
    def __init__(self, url, datastore):
        self.url = url
        self.datastore = datastore
        
        df = self.datastore.read()
        if not isinstance(df, pd.DataFrame) or df.empty:
            self.saved_id_set = set()
            self.cutoff_date = None
            self.earliest_date = None
        else:
            self.saved_id_set = set(df['id'].tolist())
            df = df.sort_values(by=['date'], ascending=False)
            self.cutoff_date = df.iloc[0].date
            self.earliest_date = df.iloc[-1].date
            
        self.new_article_list = []
        self.author_name_lookups = {}

        
    def scrape(self, total_count, count_per_page = 10):
        if total_count <= 0:
            return
        
        parameters_new = {}
        if self.cutoff_date:
            parameters_new['after'] = self.cutoff_date
            
        self._get_latest_articles(self.url, parameters_new, total_count, count_per_page)
        
        remaining_count = total_count - len(self.new_article_list)
        df = self.datastore.read()
        remaining_count -= df.shape[0]
        
        if remaining_count > 0 and self.earliest_date:
            parameters_old = {}
            parameters_old['before'] = self.earliest_date        


            print('get older articles...')
            self._get_latest_articles(self.url, parameters_old, remaining_count, count_per_page)
        
          
    def save(self):
        if len(self.new_article_list) > 0:
            self.datastore.save(pd.DataFrame(self.new_article_list))
            self.new_article_list = []
        
    def count(self):
        df = self.datastore.read()
        return len(df)

    def _get_latest_articles(self, url, parameters, total_count, count_per_page):
        page = 1
        current_total_count = total_count
        parameters['per_page'] = min(total_count, count_per_page, 100)

        while True:
            parameters['page'] = page
            response = requests.get(self.url, params = parameters) 
            response_status = response.status_code

            if response_status != 200:
                print('Warning: requests.get {} {} status: {}'.format(self.url, parameters, response_status))
                break

            try:
                response_json = response.json() 
            except ValueError as e:
                break
            else:
                if len(response_json) == 0:
                    break
                    
                for article in response_json:
                    article_id = article['id']
                    
                    if len(self.saved_id_set) == 0 or article_id not in self.saved_id_set:
                        self.saved_id_set.add(article_id)
                        self.new_article_list.append(self._parse(article))

                    current_total_count -= 1
                    if current_total_count <= 0:
                        break
                        
            finally:
                if current_total_count <= 0:
                    break
                    
                page += 1
                

    def _clean(self, raw_html):
        if isinstance(raw_html, dict) and 'rendered' in raw_html:
            raw_html = raw_html['rendered']

        clean_r = re.compile('<.*?>')
        clean_text = re.sub(clean_r, '', raw_html)
        return clean_text

    def _get_author_name(self, article):
        author_str = article['author']
        if author_str in self.author_name_lookups:
            return self.author_name_lookups[author_str]
        
        if '_links' in article:
            if 'author' in article['_links']:
                if 'href' in article['_links']['author'][0]:
                    url = article['_links']['author'][0]['href']
                    response = requests.get(url) 
                    response_status = response.status_code
                    if response_status != 200:
#                          print('Warning: cannot get author name, use author id instead: \n. {} status: {}'
#                               .format(url, response_status))
                        url = article['_links']['authors'][0]['href']
                        response = requests.get(url) 
                        response_status = response.status_code
                        if response_status != 200:
                            print('requests.get {} fails:\n  response status: {}'
                                  .format(url, response_status))
                    else:
                        try:
                            response_json = response.json() 
                        except ValueError as e:
                            print('response.json {} fails:\n  {}'.format(self.url, e))
                        else:
                            if 'name' in response.json():
                                author_str = response.json()['name']
                                self.author_name_lookups[article['author']] = author_str
        return author_str


    def _get_article_field(self, article, key):
        try:
            v = article[key]
        except KeyError:
            v = 'NA'
        finally:
            return v
            
    def _parse(self, article):
        article_id = article['id']
        author = self._get_article_field(article, 'author')
        author_name = self._get_author_name(article)
        date = self._get_article_field(article, 'date')
        title = self._clean(self._get_article_field(article, 'title'))
        link = self._get_article_field(article, 'link')
        content = self._clean(self._get_article_field(article, 'content'))
        
        return {
                'id': article_id,
                'author': author,
                'author_name': author_name,
                'date': date,
                'title': title,
                'link': link,
                'content': content
            }




import unittest
import random
import math

from wpscraper import WPScraper
from datastores import CSVDataStore, SqliteDataStore

class WPScraperTestBase(unittest.TestCase):
    def _get_articles(self, n, max_count=None):
        if n == 1:
            print('scrape latest {} article...'.format(n))
        else:
            print('scrape latest {} articles...'.format(n))

        self.scraper.scrape(n)
        self.scraper.save()
        if max_count:
            self.assertEqual(self.scraper.count(), max(max_count,n))
        else:
            self.assertEqual(self.scraper.count(), n)


    def _scrape_times(self, n, given_count=None):
        for i in range(n):
            if not given_count:
                count = math.ceil(random.random()*1000)
            else:
                count = given_count
            
            if not self.max_count:
                self.max_count = count
            else:
                self.max_count = max(self.max_count, count)
                
            self._get_articles(count,self.max_count)
        return self.max_count

        

class WPScraperTestCSV(WPScraperTestBase):
    def setUp(self):
        self.csv = CSVDataStore('techcrunch-unittesting.csv')
        self.scraper = WPScraper('http://techcrunch.com/wp-json/wp/v2/posts?', self.csv)
        self.max_count = 0


    def test_scrape_1_times_1(self):
        count = self._scrape_times(1, 1)
        self.assertEqual(self.scraper.count(), count)

    def test_scrape_3_times_18_1_37(self):
        self._scrape_times(1, 18)
        self._scrape_times(1, 1)
        self._scrape_times(1, 37)
        self.assertEqual(self.scraper.count(), 37)

    # def test_scrape_3_times_1000(self):
    #     count = self._scrape_times(3, 1000)
    #     self.assertEqual(self.scraper.count(), count)

    # def test_scrape_100_times_random(self):
    #     count = self._scrape_times(100)
    #     self.assertEqual(self.scraper.count(), count)

#     def test_scrape_3_times_418_140_937(self):
#         self._scrape_times(1, 418)
#         self._scrape_times(1, 140)
#         self._scrape_times(1, 937)
#         self.assertEqual(self.scraper.count(), 937)



    def tearDown(self):
        self.csv.clean()

            
class WPScraperTestDB(WPScraperTestBase):
    def setUp(self):
        self.db = SqliteDataStore('techcrunch.db', 'techcrunch')
        self.scraper = WPScraper('http://techcrunch.com/wp-json/wp/v2/posts?', self.db)
        self.max_count = 0
        
    def test_scrape_1_times_1(self):
        count = self._scrape_times(1, 1)
        self.assertEqual(self.scraper.count(), count)

    def test_scrape_3_times_18_1_37(self):
        self._scrape_times(1, 18)
        self._scrape_times(1, 1)
        self._scrape_times(1, 37)
        self.assertEqual(self.scraper.count(), 37)

    # def test_scrape_3_times_random(self):
    #     count = self._scrape_times(3)
    #     self.assertEqual(self.scraper.count(), count)

    # def test_scrape_3_times_101(self):
    #     count = self._scrape_times(3, 101)
    #     self.assertEqual(self.scraper.count(), count)

    # def test_scrape_3_times_1000(self):
    #     count = self._scrape_times(3, 1000)
    #     self.assertEqual(self.scraper.count(), count)

    # def test_scrape_100_times_random(self):
    #     count = self._scrape_times(100)
    #     self.assertEqual(self.scraper.count(), count)

#     def test_scrape_1_times_909(self):
#         count = self._scrape_times(1, 909)
#         self.assertEqual(self.scraper.count(), count)

#     def test_scrape_2_times_903_909(self):
#         count = self._scrape_times(1, 903)
#         count = self._scrape_times(1, 909)
#         self.assertEqual(self.scraper.count(), 909)

#     def test_scrape_2_times_903_909(self):
#         count = self._scrape_times(1, 903)
#         count = self._scrape_times(1, 909)
#         self.assertEqual(self.scraper.count(), 909)


    def tearDown(self):
        self.db.clean()





from setuptools import setup

setup(name='wpscraper',
      version='0.1',
      description='WordPress Site Recent Articles Scraper',
      packages=['wpscraper'],
      install_requires=[
            'requests',
            'pandas',
            'sqlalchemy',
      ],
      author='Faye Hall',
      author_email='fayedaihall@mac.com',
      url='https://github.com/fayehall',
      zip_safe=False)
